from typing import Callable, Any, Optional, Union
from typing import List, TypedDict
from uuid import uuid4

import ipywidgets
import orjson
from IPython.core.display import Javascript
from IPython.core.display import display
from ipywidgets import HTML

from sm_widgets.widgets.base import BaseWidget


class _Slider(BaseWidget):
    pass


class Slider(_Slider):
    def __init__(self, app: BaseWidget, app_render_fn: Union[Callable, str], dev: bool = False, embedded_error_log: bool = False):
        super().__init__("slider", dev=dev, embedded_error_log=embedded_error_log)
        self.app = app
        self.app_render_fn = getattr(app, app_render_fn) if isinstance(app_render_fn, str) else app_render_fn
        self.index = 0
    
    def render(self, same_tab: bool = True, new_window: bool = False, shadow_dom: Optional[bool] = None):
        if shadow_dom is None:
            shadow_dom = True if same_tab else False

        setup_fn = 'setupApp' + self.app_id.replace("-", "")
        self.app.render(same_tab, new_window, shadow_dom)
        widgets = [self.tunnel, *self.error_container]
        widgets.append(Javascript(f"""
            {self.RepeatUntilSuccess}
            function {setup_fn}() {{
                if (window.IPyCallback === undefined || window.IPyApps === undefined) {{
                    return false;
                }}
                let tunnel = window.IPyCallback.get('{self.tunnel.tunnel_id}');
                if (tunnel === undefined) {{
                    return false;
                }}
                
                let appwin = window.IPyApps.get('{self.app.app_id}');
                if (appwin === undefined) {{
                    return false;
                }}
                
                let container = appwin.document.getElementById('{self.app.app_id}');
                if (container === null) {{
                    return false;
                }}
                
                let div = appwin.document.createElement("div");
                div.id = '{self.app_id}';
                div.style = "margin-bottom: 8px";
                container.parentElement.prepend(div);
                
                // use the tunnel first to send out the code, after the application is rendered, the listening function 
                // is going to be replaced by the listener in the application, so we don't have to worry.
                tunnel.on_receive(function (version, msg) {{
                    let payload = JSON.parse(msg);
                    if (payload.id !== '/init/download_code') {{
                        alert('invalid calling order. you need to set the source code first');
                        console.error("invalid call order", payload);
                        return;
                    }}
                    appwin.eval(payload.response);
                    appwin.{self.app_js_render_fn}('{self.app_id}', tunnel);
                    window.IPyApps.set('{self.app_id}', appwin);
                    tunnel.send_msg(JSON.stringify({{ url: '/init/done', params: null, id: '/init/done' }}));
                }});
                tunnel.send_msg(JSON.stringify({{ url: '/init/download_code', params: null, id: '/init/download_code' }}));
                return true;
            }}
            repeatUntilSuccess({setup_fn}, 50, 10);
        """))
        display(*widgets)
        return self
    
    def set_data(self, lst: List[TypedDict('SliderAppLst', description=str, args=Any)], start_index: int=0):
        self.wait_for_app_ready(5)
        self.lst = lst
        self.index = start_index
        self.tunnel.send_msg(orjson.dumps([
            {
                "type": "wait_for_client_ready"
            },
            {
                "type": "set_props",
                "props": {
                    "min": 0,
                    "max": len(lst) - 1,
                    "index": self.index,
                    "description": lst[self.index]['description']
                }
            }
        ]).decode())
        self.app_render_fn(*lst[self.index]['args'])

    @_Slider.register_handler("/view")
    def view(self, params: dict):
        self.index = params['index']
        self.tunnel.send_msg(orjson.dumps([
            {
                "type": "set_props",
                "props": {
                    "description": self.lst[self.index]['description']
                }
            }
        ]).decode())
        self.app_render_fn(*self.lst[self.index]['args'])


class DisplayShell:
    """A basic application to make default jupyter application work with slider"""
    def __init__(self, render_fn: Callable[[Any], None]):
        self.output = ipywidgets.Output()
        self.render_fn = render_fn
        self.app_id = str(uuid4())

    def render(self, same_tab: bool = True, new_window: bool = False, shadow_dom: Optional[bool] = None):
        assert same_tab is True and new_window is False
        display(HTML(f'<div id="{self.app_id}"></div>'), self.output)
        display(Javascript(f"""
        if (window.IPyApps === undefined) {{
            window.IPyApps = new Map();
        }}
        window.IPyApps.set('{self.app_id}', window);
        """))

    def set_data(self, item):
        with self.output:
            self.output.clear_output()
            self.render_fn(item)



import enum
from dataclasses import dataclass
from typing import Dict, List, Union, Optional


class ValueType(enum.Enum):
    URI = "uri"
    String = "string"
    Integer = "integer"
    Float = "float"


@dataclass
class Value:
    __slots__ = ('type', 'value')
    # uri, string, integer, float
    type: ValueType
    value: Union[str, int, float]

    def is_uri(self):
        return self.type == ValueType.URI

    def as_uri(self):
        if self.type == ValueType.URI:
            return self.value
        assert f"Cannot convert value of type {self.type} to URI"


@dataclass
class Statement:
    value: Value
    qualifiers: Dict[str, List[Value]]


@dataclass
class Entity:
    # id or uri of the entity
    uri: str
    label: str
    description: str
    properties: Dict[str, List[Statement]]

    @property
    def readable_label(self):
        return self.label





from dataclasses import dataclass

from typing import Optional, TypedDict, Union, List, Tuple

from sm.misc import OntNS

from sm_widgets.models import Table

Resource = TypedDict("Resource", dict(uri=str, label=str))


@dataclass
class ColumnRelationship:
    # dict for entity, number for column index
    endpoint: Union[Resource, int]
    predicates: Tuple[Resource, Resource]
    freq: int


ColumnRelationshipResult = TypedDict("ColumnRelationshipResult",
                                     dict(incoming=List[ColumnRelationship], outgoing=List[ColumnRelationship]))


# noinspection PyMethodMayBeStatic
class AnnotatorAssistant:

    def get_row_indices(self, table: Table, source_node: Union[int, str], target_node: Union[int, str],
                        links: Tuple[str, str]):
        # return the list of indices
        return set()

    def get_column_relationships(self, table: Table, column_index: str) -> Optional[ColumnRelationshipResult]:
        # get relationships of columns, should return {
        # "incoming": {endpoint, endpoint_type, predicates, freq}[]
        # "outgoing": {endpoint, endpoint_type, predicates, freq}[]
        # }
        return None


class DummyAnnotatorAssistant(AnnotatorAssistant):

    def __init__(self):
        self.ontns = OntNS.get_instance()

    def get_column_relationships(self, table: Table, column_index: str) -> Optional[ColumnRelationshipResult]:
        return {
            "incoming": [
                ColumnRelationship(
                    endpoint=0,
                    predicates=(
                        {"uri": self.ontns.get_abs_uri("p:P585"), "label": "Point in time (P585)"},),
                    freq=10
                ),
                ColumnRelationship(
                    endpoint=0,
                    predicates=({"uri": self.ontns.get_abs_uri("p:P710"), "label": "participant (P710)"},),
                    freq=5
                ),
                ColumnRelationship(
                    endpoint={"uri": self.ontns.get_abs_uri("wd:Q5"), "label": "United States (Q5)"},
                    predicates=({"uri": self.ontns.get_abs_uri("p:P585"), "label": "Point in time (P585)"},),
                    freq=10
                )
            ],
            "outgoing": [
                ColumnRelationship(
                    endpoint=0,
                    predicates=({"uri": self.ontns.get_abs_uri("p:P585"), "label": "Point in time (P585)"},),
                    freq=10
                ),
                ColumnRelationship(
                    endpoint=0,
                    predicates=({"uri": self.ontns.get_abs_uri("p:P710"), "label": "participant (P710)"},),
                    freq=5
                ),
                ColumnRelationship(
                    endpoint={"uri": self.ontns.get_abs_uri("wd:Q5"), "label": "United States (Q5)"},
                    predicates=({"uri": self.ontns.get_abs_uri("p:P585"), "label": "Point in time (P585)"},),
                    freq=10
                )
            ]
        }





from dataclasses import dataclass, field
from typing import List, Set, Optional


@dataclass
class OntClass:
    uri: str
    label: str
    aliases: List[str]
    description: str
    parents: List[str]
    parents_closure: Set[str] = field(default_factory=set)

    @property
    def readable_label(self):
        return self.label


@dataclass
class OntProperty:
    uri: str
    label: str
    aliases: List[str]
    description: str
    parents: List[str]
    parents_closure: Set[str] = field(default_factory=set)

    @property
    def readable_label(self):
        return self.label


from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

import orjson

from sm.inputs import ColumnBasedTable, Column
import sm.misc as M
from sm_widgets.models.entity import Value


@dataclass
class ContextLevel:
    level: int
    header: str
    content: str

    def clone(self):
        return ContextLevel(self.level, self.header, self.content)


@dataclass
class Link:
    start: int
    end: int
    url: str
    # uri of entity
    entity: Optional[str]


@dataclass
class Table:
    table: ColumnBasedTable
    # list of values in the context
    context_values: List[Value]
    context_tree: List[ContextLevel]

    links: List[List[List[Link]]]

    @property
    def id(self):
        return self.table.table_id

    def size(self):
        if len(self.table.columns) == 0:
            return 0
        return len(self.table.columns[0].values)

    @staticmethod
    def from_csv_file(infile: Union[Path, str], first_row_header: bool = True, table_id: Optional[str] = None):
        infile = Path(infile)
        link_file = infile.parent / f"{infile.stem}.links.tsv"

        if table_id is None:
            table_id = infile.stem
        rows = M.deserialize_csv(infile)

        assert len(rows) > 0, "Empty table"
        columns = []
        if first_row_header:
            headers = rows[0]
            rows = rows[1:]
        else:
            headers = [f'column-{i:03d}' for i in range(len(rows[0]))]

        for ci, cname in enumerate(headers):
            columns.append(Column(ci, cname, [r[ci] for r in rows]))
        table = ColumnBasedTable(table_id, columns)
        links = []
        for ri in range(len(rows)):
            links.append([[] for ci in range(len(headers))])

        if link_file.exists():
            for row in M.deserialize_csv(link_file, delimiter="\t"):
                ri, ci, ents = int(row[0]), int(row[1]), row[2:]
                for ent in ents:
                    if ent.startswith("{"):
                        # it's json, encoding the hyperlinks
                        link = Link(**orjson.loads(ent))
                    else:
                        link = Link(0, len(table.columns[ci][ri]), ent, ent)
                    links[ri][ci].append(link)
        return Table(table, context_values=[], context_tree=[], links=links)



import functools
from typing import Callable, Any, TypeVar, Dict

K = TypeVar("K")
V = TypeVar("V")


class StoreWrapper:
    def __init__(self, store: dict, key_deser: Callable[[K], Any], val_deser: Callable[[Any], V]):
        self.store = store
        self.key_deser = key_deser
        self.val_deser = val_deser

    @functools.lru_cache
    def __contains__(self, key):
        return self.key_deser(key) in self.store

    @functools.lru_cache
    def __getitem__(self, key):
        val = self.store[self.key_deser(key)]
        return self.val_deser(val)

    def __setitem__(self, key, val):
        assert False

    def __delitem__(self, key):
        assert False

    def __len__(self):
        assert False

    @functools.lru_cache
    def get(self, key, default=None):
        key = self.key_deser(key)
        if key not in self.store:
            return default
        val = self.store[key]
        return self.val_deser(val)


from typing import List, Dict

from elasticsearch import Elasticsearch, helpers
import requests
from tqdm.auto import tqdm


class ElasticSearchService:
    def __init__(self, eshost: str, index: str):
        self.eshost = eshost
        self.index = index
        self.fields = ['id', 'label', 'uri', 'readable_label', 'aliases', 'description']
        self.search_fields = ['id^10', 'label^5', 'aliases^3', 'description']

    def search(self, query, top_k: int = 20):
        resp = requests.post(f"{self.eshost}/{self.index}/_search", json={
            "query": {
                "multi_match": {
                    "query": query,
                    "fields": self.search_fields
                }
            }
        })
        assert resp.status_code == 200, resp.status_code
        data = resp.json()
        docs = []
        for r in data['hits']['hits'][:top_k]:
            node = r['_source']
            docs.append({
                "uri": node['uri'],
                "label": node['readable_label'],
                "description": node['description'],
                "score": r['_score']
            })
        return docs

    def load2index(self, documents: List[dict], batch_size: int = 128):
        self.create_index()
        es = Elasticsearch([self.eshost])
        docs = [
            dict(_id=doc['id'], **{k: doc[k] for k in self.fields})
            for doc in documents
        ]
        for i in tqdm(range(0, len(docs), batch_size)):
            helpers.bulk(es, docs[i:i + batch_size], index=self.index)

    def create_index(self):
        index_settings = {
            "settings": {
                "analysis": {
                    "analyzer": {
                        "keyword_analyzer": {
                            "tokenizer": "standard",
                            "filter": ["lowercase"]
                        },
                        "default": {
                            "tokenizer": "standard",
                            "filter": ["lowercase", "2_16_edgegrams"]
                        },
                        "default_search": {
                            "tokenizer": "standard",
                            "filter": ["lowercase", "16_truncate"]
                        },
                    },
                    "filter": {
                        "2_16_edgegrams": {
                            "type": "edge_ngram",
                            "min_gram": 2,
                            "max_gram": 16
                        },
                        "16_truncate": {
                            "type": "truncate",
                            "length": 16
                        }
                    }
                }
            },
            "mappings": {
                "properties": {
                    "id": {
                        "type": "text",
                        "analyzer": "keyword_analyzer",
                    }
                }
            }
        }

        resp = requests.put(f"{self.eshost}/{self.index}", json=index_settings)
        assert resp.status_code == 200, resp.text


class OntologyClassSearch(ElasticSearchService):
    def __init__(self, eshost: str):
        super().__init__(eshost, "ontology-classes")


class OntologyPropertySearch(ElasticSearchService):
    def __init__(self, eshost: str):
        super().__init__(eshost, "ontology-properties")





from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from operator import attrgetter
from pathlib import Path
from typing import Dict, List, Set, Optional, Tuple, Union

import orjson
import sm.misc as M
import sm.outputs as O

from sm.misc import OntNS
from sm_widgets.models import Entity, OntClass, OntProperty, Table, Value
from sm_widgets.services.search import OntologyClassSearch, OntologyPropertySearch
from sm_widgets.widgets.annotator.assistant import AnnotatorAssistant
from sm_widgets.widgets.annotator.session import Session
from sm_widgets.widgets.annotator.tree_ordering import reorder2tree, IndirectDictAccess
from sm_widgets.widgets.base import BaseWidget
from sm_widgets.widgets.slider import Slider


class FilterOp(Enum):
    INCLUDE = "include"
    EXCLUDE = "exclude"


@dataclass
class EntityTypeFilter:
    column_index: int
    # qnode that is the class
    class_uri: str
    operator: FilterOp


@dataclass
class RelFilter:
    # the column we want to filter
    column_index: int
    # the endpoint, source or target or wildcard (both none): string for entity & int for column index
    source_endpoint: Optional[Union[int, str]]
    target_endpoint: Optional[Union[int, str]]
    # [0] is property uri, [1] is qualifier uri
    props: Tuple[str, str]
    operator: FilterOp


class CellTypeFilterOp(Enum):
    HAS_LINK = "hasLink"
    HAS_ENTITY = "hasEntity"
    HAS_NO_LINK = "noLink"
    HAS_NO_ENTITY = "noEntity"


@dataclass
class CellTypeFilter:
    column_index: int
    op: CellTypeFilterOp


class _Annotator(BaseWidget):
    pass


class Annotator(_Annotator):
    def __init__(self,
                 entities: Dict[str, Entity], ontclasses: Dict[str, OntClass], ontprops: Dict[str, OntProperty],
                 prop_type_uri: str, savedir: str, eshost: str,
                 dev: bool = False, assistant: Optional[AnnotatorAssistant] = None, embedded_error_log: bool = False):
        super().__init__("annotator", dev, embedded_error_log=embedded_error_log)

        self.savedir = Path(savedir)
        self.savedir.mkdir(exist_ok=True, parents=True)

        self.entities = entities
        self.ontclasses = ontclasses
        self.ontprops = ontprops

        self.ont_class_search = OntologyClassSearch(eshost)
        self.ont_prop_search = OntologyPropertySearch(eshost)
        self.ontns = OntNS.get_instance()
        self.assistant = assistant
        self.prop_type_uri = prop_type_uri

        self.wdclass_parents: Dict[str, Set[str]] = IndirectDictAccess(self.ontclasses, attrgetter("parents_closure"))
        self.cache_id2label: Dict[str, str] = {}

    def set_data(self, id: str, tbl: Table, sms: List[O.SemanticModel]=None):
        self.wait_for_app_ready(5)
        self.cache_id2label = {}
        infile = M.get_latest_path(self.savedir / id / "version.json")
        if infile is None:
            sms = sms or []
            is_curated = False
            note = ""
        else:
            data = M.deserialize_json(infile)
            assert data['version'] == 2
            sms = [O.SemanticModel.from_dict(sm) for sm in data['semantic_models']]
            is_curated = data['is_curated']
            note = data['note']

        self.session = Session(id, is_curated, note, tbl, sms)
        if len(self.session.graphs) == 0:
            self.session.graphs.append(O.SemanticModel())
        self.session.graphs = [self.add_default_nodes_to_sm(sm) for sm in self.session.graphs]

        self.tunnel.send_msg(orjson.dumps([
            {
                "type": "wait_for_client_ready",
            },
            {
                "type": "set_props",
                "props": {
                    "log": {
                        "isCurated": self.session.is_curated,
                        "note": self.session.note,
                    },
                    "table": self.serialize_table_schema(),
                    "graphs": [self.serialize_sm(graph) for graph in self.session.graphs],
                    "entities": {},
                    "assistant": {
                        "id": self.session.table.id
                    },
                    "currentGraphIndex": 0,
                    "wdOntology": {
                        "username": '',
                        "password": ''
                    },
                }
            },
            {
                "type": "exec_func",
                "func": "app.tableFetchData",
                "args": []
            }
        ]).decode())

    def save_annotation(self):
        assert len(self.session.graphs) > 0
        (self.savedir / self.session.id).mkdir(exist_ok=True)
        outfile = M.get_incremental_path(self.savedir / self.session.id / "version.json")
        with open(outfile, "wb") as f:
            f.write(orjson.dumps(self.session.to_json(), option=orjson.OPT_INDENT_2))

    @_Annotator.register_handler("/table")
    def fetch_table_data(self, params: dict):
        table = self.session.table
        start = params['offset']
        end = start + params['limit']
        filters = [
            EntityTypeFilter(
                item['columnId'],
                item['uri'],
                FilterOp(item['op']))
            for item in params['typeFilters']
        ]
        filters += [
            RelFilter(
                int(item['columnId']),
                (item['endpoint'] if isinstance(item['endpoint'], str) else item['endpoint']) if
                    item['direction'] == 'incoming' and item['endpoint'] != "*" else None,
                (item['endpoint'] if isinstance(item['endpoint'], str) else item['endpoint']) if
                    item['direction'] == 'outgoing' and item['endpoint'] != "*" else None,
                (item['pred1'], item['pred2']),
                FilterOp(item['op'])
            )
            for item in params['relFilters']
        ]
        filters += [
            CellTypeFilter(int(column_id), CellTypeFilterOp(op))
            for column_id, op in params['linkFilters'].items()
        ]

        tbl_size = table.size()
        if len(filters) == 0:
            total = tbl_size
            row_index = range(start, min(total, end))
        else:
            includes = set()
            excludes = set()
            for filter in filters:
                ci = filter.column_index

                if isinstance(filter, CellTypeFilter):
                    for ri in range(table.size()):
                        links = table.links[ri][ci]
                        if filter.op == CellTypeFilterOp.HAS_LINK:
                            if len(links) == 0:
                                excludes.add(ri)
                        elif filter.op == CellTypeFilterOp.HAS_NO_LINK:
                            if len(links) > 0:
                                excludes.add(ri)
                        elif filter.op == CellTypeFilterOp.HAS_ENTITY:
                            if all(link.entity is None for link in links):
                                excludes.add(ri)
                        else:
                            assert filter.op == CellTypeFilterOp.HAS_NO_ENTITY
                            if not all(link.entity is None for link in links):
                                excludes.add(ri)
                    continue

                if filter.operator == FilterOp.EXCLUDE:
                    sat_condition = excludes
                else:
                    sat_condition = includes

                if isinstance(filter, RelFilter):
                    if filter.source_endpoint is not None:
                        assert filter.target_endpoint is None
                        # query for source
                        res = self.assistant.get_row_indices(table, filter.source_endpoint, filter.column_index,
                                                             filter.props)
                    elif filter.target_endpoint is not None:
                        assert filter.source_endpoint is None
                        # query for the target
                        res = self.assistant.get_row_indices(table, filter.column_index, filter.target_endpoint,
                                                             filter.props)
                    else:
                        res = set()
                        for ri in range(table.size()):
                            links = table.links[ri][ci]
                            sat = False
                            for link in links:
                                if link.entity is None:
                                    continue
                                qnode = self.entities[link.entity]
                                if filter.props[0] not in qnode.properties:
                                    continue
                                if filter.props[0] == filter.props[1]:
                                    # statement value
                                    sat = True
                                else:
                                    for stmt in qnode.properties[filter.props[0]]:
                                        if filter.props[1] in stmt.qualifiers:
                                            sat = True
                                            break
                                if sat:
                                    break
                            if sat:
                                res.add(ri)
                    sat_condition.update(res)
                else:
                    assert isinstance(filter, EntityTypeFilter)
                    for ri in range(table.size()):
                        links = table.links[ri][ci]
                        sat = False
                        for link in links:
                            if link.entity is None:
                                continue
                            qnode = self.entities[link.entity]
                            for stmt in qnode.properties.get(self.prop_type_uri, []):
                                classid = stmt.value.as_uri()
                                if classid not in self.ontclasses:
                                    continue
                                clsnode = self.ontclasses[classid]
                                # this entity is the same class, or its type is child of the desired class not parents
                                if classid == filter.class_uri or filter.class_uri in clsnode.parents_closure:
                                    sat = True
                                    break
                            if sat:
                                break
                        if sat:
                            sat_condition.add(ri)

            if len(includes) == 0 and len(excludes) > 0:
                includes = self.session.table_records_index
            row_index = sorted(includes.difference(excludes))
            total = len(row_index)
            row_index = row_index[start:min(total, end)]

        rows = []
        for ri in row_index:
            row = []
            for ci in range(len(table.table.columns)):
                row.append(self.serialize_table_cell(ri, ci))

            rows.append({
                "data": row,
                "rowId": ri
            })
        return {"rows": rows, "total": total}

    @_Annotator.register_handler("/entities")
    def get_entity(self, params: dict):
        return {
            uri: self.serialize_entity(uri, full=True)
            for uri in params['uris']
        }
    
    @_Annotator.register_handler("/ontology/class")
    def search_ont_class(self, params: dict):
        return self.ont_class_search.search(params['query'])

    @_Annotator.register_handler("/ontology/predicate")
    def search_ont_property(self, params: dict):
        return self.ont_prop_search.search(params['query'])

    @_Annotator.register_handler("/save")
    def save_annotation_handler(self, params: dict):
        self.session.note = params['note']
        self.session.is_curated = params['isCurated']
        self.session.graphs = [self.deserialize_sm(g['nodes'], g['edges']) for g in params['graphs']]
        self.save_annotation()

    @_Annotator.register_handler("/assistant/column")
    def suggest_column(self, params: dict):
        ci = params['columnIndex']
        table = self.session.table
        column = table.table.columns[ci]

        # compute statistics
        stats = {
            'entities/linked row': [],
            "links/row": [],
            "avg %link surface": [],
            "# rows": table.size(),
        }
        qnode2types = {}
        for ri, value in enumerate(column.values):
            nchars = len(value)
            links = table.links[ri][ci]
            n_covered_chars = sum(l.end - l.start for l in links)
            n_qnodes = sum(l.entity is not None for l in links)
            if len(links) > 0:
                stats['entities/linked row'].append(n_qnodes)
            stats['links/row'].append(len(links))
            stats['avg %link surface'].append(n_covered_chars / max(nchars, 1e-7))

            for l in links:
                if l.entity is None:
                    continue
                qnode = self.entities[l.entity]
                qnode2types[l.entity] = list({stmt.value.as_uri() for stmt in qnode.properties.get(self.prop_type_uri, [])})

        stats['# unique qnodes'] = len(qnode2types)
        for k, v in stats.items():
            if isinstance(v, list):
                if len(v) == 0:
                    stats[k] = "0%"
                else:
                    stats[k] = f"{sum(v) / max(len(v), 1e-7):.2f}%"

        supertype_freq = defaultdict(int)

        for lst in qnode2types.values():
            etypes = set()
            for item in lst:
                etypes.add(item)
                etypes.update(self.ontclasses[item].parents_closure)
            for c in etypes:
                supertype_freq[c] += 1

        flat_type_hierarchy = reorder2tree(list(supertype_freq.keys()), self.wdclass_parents)
        qnode2children = {}

        def traversal(tree, path):
            if tree.id not in qnode2children:
                qnode2children[tree.id] = set()
            for parent_tree in path:
                qnode2children[parent_tree.id].add(tree.id)

        flat_type_hierarchy.preorder(traversal)

        flat_type_hierarchy = flat_type_hierarchy.update_score(lambda n: supertype_freq[n.id]).sort(reverse=True)
        flat_type_hierarchy = [
            {
                "uri": u.id,
                "label": self.get_entity_label(u.id),
                "duplicated": u.duplicated,
                "depth": u.depth,
                "freq": supertype_freq[u.id],
                "percentage": supertype_freq[u.id] / len(qnode2types)
            }
            for u in flat_type_hierarchy.get_flatten_hierarchy(dedup=True)
        ]

        resp = {
            "stats": stats,
            "flattenTypeHierarchy": flat_type_hierarchy,
            "type2children": {k: list(v) for k, v in qnode2children.items()},
        }

        if self.assistant is not None:
            relationships = self.assistant.get_column_relationships(table, ci)
            if relationships is not None:
                for rels in relationships.values():
                    for i, rel in enumerate(rels):
                        rels[i] = dict(
                            endpoint=rel.endpoint,
                            predicates=rel.predicates,
                            freq=rel.freq,
                        )
                resp['relationships'] = relationships
        return resp

    def add_default_nodes_to_sm(self, sm: O.SemanticModel):
        table = self.session.table
        column2name = self.session.column2name
        for col in table.table.columns:
            if not any(n.is_data_node and n.col_index == col.index for n in sm.iter_nodes()):
                # no column in the model, add missing columns
                dnodeid = f'd-{col.index}'
                assert not sm.has_node(dnodeid)
                sm.add_node(O.DataNode(f"d-{col.index}", col.index, column2name[col.index]))

        # if table.context.page_qnode is not None:
        #     context_nodeid = f"context-{table.context.page_qnode}"
        #     sm.add_node(O.LiteralNode(
        #         context_nodeid,
        #         WDOnt.get_qnode_uri(table.context.page_qnode),
        #         f"{self.entities[table.context.page_qnode].label} ({table.context.page_qnode})",
        #         True,
        #         O.LiteralNodeDataType.Entity
        #     ))
        return sm

    def deserialize_sm(self, nodes: List[dict], edges: List[dict]) -> O.SemanticModel:
        sm = O.SemanticModel()
        for n in nodes:
            if n['isDataNode']:
                sm.add_node(O.DataNode(n['id'], n['columnId'], self.session.column2name[n['columnId']]))
            elif n['isClassNode']:
                rel_uri = self.ontns.get_rel_uri(n['uri'])
                sm.add_node(
                    O.ClassNode(n['id'], n['uri'], rel_uri, n['approximation']))
            else:
                assert n['isLiteralNode']
                sm.add_node(
                    O.LiteralNode(n['id'], n['uri'], n['label'], n['isInContext'], O.LiteralNodeDataType(n['datatype'])))

        for e in edges:
            rel_uri = self.ontns.get_rel_uri(e['uri'])
            sm.add_edge(O.Edge(e['source'], e['target'], e['uri'], rel_uri, e['approximation']))
        return sm

    def serialize_sm(self, sm: O.SemanticModel):
        nodes = []
        for n in sm.iter_nodes():
            if n.is_class_node:
                nodes.append({
                    "id": n.id,
                    "uri": n.abs_uri,
                    "label": self.get_entity_label(n.abs_uri) or n.rel_uri,
                    "approximation": n.approximation,
                    "isClassNode": True,
                    "isDataNode": False,
                    "isLiteralNode": False,
                })
            elif n.is_data_node:
                nodes.append({
                    "id": n.id,
                    "label": self.session.column2name[n.col_index],
                    "isClassNode": False,
                    "isDataNode": True,
                    "isLiteralNode": False,
                    "columnId": n.col_index,
                })
            else:
                nodes.append({
                    "id": n.id,
                    "uri": n.value if n.datatype == O.LiteralNodeDataType.Entity else "",
                    "label": n.label,
                    "isClassNode": False,
                    "isDataNode": False,
                    "isLiteralNode": True,
                    "isInContext": n.is_in_context,
                    "datatype": n.datatype.value
                })

        return {
            "tableID": self.session.table.id,
            "nodes": nodes,
            "edges": [
                {
                    "source": e.source,
                    "target": e.target,
                    "uri": e.abs_uri,
                    "label": self.get_entity_label(e.abs_uri) or e.rel_uri,
                    "approximation": e.approximation
                }
                for e in sm.iter_edges()
            ]
        }

    def serialize_table_schema(self):
        """Get schema of the table"""
        table = self.session.table
        columns = [{"title": "", "dataIndex": "rowId"}]
        for col in table.table.columns:
            columns.append({
                "title": col.name,
                "columnId": col.index,
                "dataIndex": ["data", col.index],
            })
        return {
            "id": table.id,
            "columns": columns,
            "rowKey": "rowId",
            "totalRecords": table.size(),
            "metadata": {
                # "title": table.context.page_title,
                # "url": table.context.page_url,
                # "entity": {"uri": WDOnt.get_qnode_uri(table.context.page_qnode), "label": self.get_entity_label(table.context.page_qnode)} if table.context.page_qnode is not None else None
                "title": "",
                "url": "",
                "entity": None
            }
        }

    def serialize_table_cell(self, ri: int, ci: int):
        table = self.session.table
        value = table.table.columns[ci].values[ri]
        qnodes_metadata = {}
        for link in table.links[ri][ci]:
            if link.entity is not None:
                qnodes_metadata[link.entity] = self.serialize_entity(link.entity)

        return {
            "value": value,
            "links": [
                {
                    "start": link.start, "end": link.end,
                    "href": link.url,
                    "entity": link.entity
                }
                for link in table.links[ri][ci]
            ],
            "metadata": {
                "entities": qnodes_metadata
            }
        }

    def serialize_entity(self, uri: str, full: bool = False):
        wdclass_parents = self.wdclass_parents
        qnode = self.entities[uri]
        ent = {
            "uri": uri,
            "label": str(self.get_entity_label(uri)),
        }
        try:
            # get hierarchy
            if self.prop_type_uri in qnode.properties:
                forest = reorder2tree(
                    [stmt.value.as_uri() for stmt in qnode.properties.get(self.prop_type_uri, [])],
                    wdclass_parents)
                hierarchy = [{"uri": x.id,
                              "label": self.get_entity_label(x.id),
                              "depth": x.depth}
                             for x in forest.get_flatten_hierarchy()]
            else:
                hierarchy = []

            ent["types"] = hierarchy
            if not full:
                return ent

            props = {}
            for puri, stmts in qnode.properties.items():
                ser_stmts = []
                for stmt in stmts:
                    ser_stmts.append({
                        "value": self.serialize_datavalue(stmt.value),
                        "qualifiers": {
                            qid: {
                                "uri": qid,
                                "label": self.get_entity_label(qid),
                                "values": [self.serialize_datavalue(qual) for qual in quals]
                            }
                            for qid, quals in stmt.qualifiers.items()
                        }
                    })

                props[puri] = {
                    "uri": puri,
                    "label": self.get_entity_label(puri),
                    "values": ser_stmts
                }
            ent['props'] = props
            ent['description'] = str(qnode.description)
        except KeyError:
            print(f"Error while obtaining information of a qnode: {uri}")
            raise

        return ent

    def serialize_datavalue(self, val: Value):
        if val.is_uri():
            uri = val.as_uri()
            # an exception that they sometime has property instead of qnode id???
            # this is probably an error, but we need to make this code robust
            label = self.get_entity_label(uri)
            return {"uri": uri, "label": label}
        return val.value

    def get_entity_label(self, uri: str):
        if uri not in self.cache_id2label:
            if uri in self.entities:
                label = self.entities[uri].readable_label
            elif uri in self.ontclasses:
                label = self.ontclasses[uri].readable_label
            elif uri in self.ontprops:
                label = self.ontprops[uri].readable_label
            else:
                label = None

            self.cache_id2label[uri] = label
        return self.cache_id2label[uri]


class BatchAnnotator(Slider):

    def __init__(self, annotator: Annotator, dev: bool = False):
        super().__init__(annotator, annotator.set_data, dev)

    def set_data(self, tables_with_ids: List[Tuple[str, str, Table, Optional[List[O.SemanticModel]]]], start_index: int=0):
        super().set_data([
            dict(description=description, args=(table_id, table, sms))
            for table_id, description, table, sms in tables_with_ids
        ], start_index)



import os
import time
import traceback
from typing import Optional
from uuid import uuid4

import orjson
from IPython.core.display import Javascript, display
from ipycallback import SlowTunnelWidget
from ipywidgets import HTML, Output
from loguru import logger
from sm.misc import deserialize_text


class BadRequest(Exception):
    pass


class BaseWidget:
    URL2HANDLER = {
        '/init/download_code': lambda self, params: self.jscode,
        '/init/done': lambda self, params: setattr(self, 'app_is_loaded', True)
    }

    RepeatUntilSuccess = r"""
    function repeatUntilSuccess(fn, timeout, maxTry) {
        if (fn() === true) {
            return;
        }
        if (maxTry === undefined) {
            maxTry = 10;
        }
        if (maxTry === 0) {
            console.error("Max retries error");
            alert("max retries error");
            throw new Error("Max retries error");
        } else {
            setTimeout(function () {
                repeatUntilSuccess(fn, timeout, maxTry - 1);
            }, timeout);
        }
    }
    """

    def __init__(self, app_name: str, dev: bool = False, embedded_error_log: bool = False):
        self.dev = dev
        current_dir = os.path.dirname(os.path.abspath(__file__))
        if dev:
            infile = os.path.join(current_dir, f"../../webapps/{app_name}/build/static/js/main.js")
            if not os.path.exists(infile):
                infile = os.path.join(current_dir, f"jscodes/{app_name}.js.gz")
        else:
            infile = os.path.join(current_dir, f"jscodes/{app_name}.js.gz")
        self.jscode = deserialize_text(infile)
        self.app_id = str(uuid4())
        self.app_js_render_fn = "".join([s.capitalize() for s in app_name.replace("-", "_").split("_")]) + "." + ('renderDevApp' if self.dev else 'renderApp')
        self.tunnel = SlowTunnelWidget()
        self.tunnel.on_receive(self.communicate)
        self.app_is_loaded = False
        self.error_container = []
        if embedded_error_log:
            self.error_container.append(Output())

    def render(self, same_tab: bool = True, new_window: bool = False, shadow_dom: Optional[bool] = None):
        if shadow_dom is None:
            shadow_dom = True if same_tab else False

        setup_fn = 'setupApp' + self.app_id.replace("-", "")
        if same_tab:
            widgets = [self.tunnel, HTML(f'<div id="{self.app_id}"></div>'), *self.error_container]
            setup_html_container = "let win = window;"
        else:
            widgets = [self.tunnel, *self.error_container]
            if new_window:
                # important to use "about:blank", otherwise, you can't zoom
                open_window_code = f'window.open("about:blank", "app-{self.app_id}", "toolbar=no, location=no, directories=no, status=no, menubar=no, scrollbars=yes, resizable=yes, width=window.outerWidth, height=window.outerHeight, top="+(screen.height)+", left="+(screen.width)");'
            else:
                open_window_code = f'window.open("about:blank", "app-{self.app_id}");'
            setup_html_container = f"""
            let win = {open_window_code}
            let title = win.document.createElement("title");
            title.innerHTML = '{self.__class__.__name__}';
            win.document.head.appendChild(title);
            
            let div = win.document.createElement("div");
            div.id = '{self.app_id}';
            win.document.body.appendChild(div);
            win.document.body.style = "padding: 8px 8px 0 8px;";
            """

        widgets.append(Javascript(f"""
        {self.RepeatUntilSuccess}
        if (window.IPyApps === undefined) {{
            window.IPyApps = new Map();
        }}
        function {setup_fn}() {{
            // get tunnel
            if (window.IPyCallback === undefined) {{
                return false;
            }}
            let tunnel = window.IPyCallback.get('{self.tunnel.tunnel_id}');
            if (tunnel === undefined) {{
                return false;
            }}
            
            // setup html container
            {setup_html_container}
            
            // use the tunnel first to send out the code, after the application is rendered, the listening function 
            // is going to be replaced by the listener in the application, so we don't have to worry.
            tunnel.on_receive(function (version, msg) {{
                let payload = JSON.parse(msg);
                if (payload.id !== '/init/download_code') {{
                    alert('invalid calling order. you need to set the source code first');
                    console.error("invalid call order. waiting for source code but get:", payload);
                    return;
                }}
                win.eval(payload.response);
                let shadowDOM = {str(shadow_dom).lower()};
                win.{self.app_js_render_fn}('{self.app_id}', tunnel, undefined, shadowDOM);
                window.IPyApps.set('{self.app_id}', win);
                tunnel.send_msg(JSON.stringify({{ url: '/init/done', params: null, id: '/init/done' }}));
            }});
            tunnel.send_msg(JSON.stringify({{ url: '/init/download_code', params: null, id: '/init/download_code' }}));
            return true;
        }}
        repeatUntilSuccess({setup_fn}, 50, 10);
        """))
        display(*widgets)
        return self

    def communicate(self, version: int, msg: str):
        try:
            payload = orjson.loads(msg)
            request_id, url, params = payload['id'], payload['url'], payload['params']
            resp = self.URL2HANDLER[url](self, params)
            is_success = True
        except BadRequest as e:
            is_success = False
            resp = str(e)
        except Exception as e:
            # having it here since default Jupyter Notebook doesn't show the log
            if not hasattr(BaseWidget, '_has_inited_loguru_'):
                setattr(BaseWidget, '_has_inited_loguru_', True)
                logger.add("/tmp/notebook_app.log")
            logger.exception("Error while handling incoming messasges")

            if len(self.error_container) > 0:
                with self.error_container[0]:
                    traceback.print_exc()
            raise

        self.tunnel.send_msg(orjson.dumps({
            "type": "response",
            "id": request_id,
            "success": is_success,
            "response": resp
        }).decode())

    def wait_for_app_ready(self, timeout_seconds: int):
        # pulling every 100ms
        idle = 0.1
        n_pull = int(timeout_seconds / idle)
        for i in range(n_pull):
            if self.app_is_loaded:
                return True
            time.sleep(idle)
        raise Exception("Timeout")

    @classmethod
    def register_handler(cls, url: str):
        def wrapper_fn(func):
            cls.URL2HANDLER[url] = func
            return func
        return wrapper_fn



import orjson

from sm_widgets.widgets.base import BaseWidget


class FoldableList(BaseWidget):
    def __init__(self, dev: bool = False):
        super().__init__("foldable_list", dev=dev)

    def show_list(self, items: list, header: str=None):
        props = {"items": items}
        if header is not None:
            props['header'] = header
        self.tunnel.send_msg(orjson.dumps([
            {
                "type": "wait_for_client_ready",
            },
            {
                "type": "set_props",
                "props": props
            }
        ]).decode())


from dataclasses import dataclass
from pathlib import Path

from rdflib import RDFS

from sm.misc import OntNS

from kgdata.wikidata.models import QNode, WDClass, WDProperty, DataValue
import kgdata.wikidata.db as kg_db
from sm_widgets.models import Entity, Value, OntClass, OntProperty
from sm_widgets.models.entity import ValueType, Statement
from sm_widgets.services.search import OntologyClassSearch, OntologyPropertySearch
from sm_widgets_integration.common import StoreWrapper


class GramsIntFn:
    ontns = OntNS.get_instance()

    @dataclass
    class WrapperQNode(Entity):
        id: str

        @property
        def readable_label(self):
            return f"{self.label} ({self.id})"

    @dataclass
    class WrapperWDClass(OntClass):
        id: str = None

        @property
        def readable_label(self):
            return f"{self.label} ({self.id})"

    @dataclass
    class WrapperWDProperty(OntProperty):
        id: str = None

        @property
        def readable_label(self):
            return f"{self.label} ({self.id})"

    @staticmethod
    def key_deser(uri: str):
        if uri.startswith("http://www.wikidata.org/"):
            uri = uri.replace("http://www.wikidata.org/entity/", "")
            uri = uri.replace("http://www.wikidata.org/prop/", "")
        elif uri.startswith("http://wikidata.org/"):
            uri = uri.replace("http://wikidata.org/entity/", "")
            uri = uri.replace("http://wikidata.org/prop/", "")
        return uri

    @staticmethod
    def qnode_deser(qnode: QNode):
        props = {}
        for pid, stmts in qnode.props.items():
            new_stmts = []
            for stmt in stmts:
                new_stmt = Statement(value=GramsIntFn.wd_value_deser(stmt.value),
                                     qualifiers={
                                         f'http://www.wikidata.org/prop/{qid}': [GramsIntFn.wd_value_deser(x) for x in lst]
                                         for qid, lst in stmt.qualifiers.items()
                                     })
                new_stmts.append(new_stmt)
            props[f'http://www.wikidata.org/prop/{pid}'] = new_stmts
        return GramsIntFn.WrapperQNode(
            id=qnode.id,
            uri=f"http://www.wikidata.org/entity/{qnode.id}",
            label=qnode.label,
            description=qnode.description,
            properties=props
        )

    @staticmethod
    def ont_class_deser(item: WDClass):
        parents = [f"http://www.wikidata.org/entity/{p}" for p in item.parents]
        parents_closure = {
            f"http://www.wikidata.org/entity/{p}" for p in item.parents_closure
        }
        return GramsIntFn.WrapperWDClass(id=item.id, uri=item.get_uri(), aliases=item.aliases,
                                         label=item.label, description=item.description,
                                         parents=parents, parents_closure=parents_closure)

    @staticmethod
    def ont_prop_deser(item: WDProperty):
        parents = [f"http://www.wikidata.org/entity/{p}" for p in item.parents]
        parents_closure = {
            f"http://www.wikidata.org/entity/{p}" for p in item.parents_closure
        }
        return GramsIntFn.WrapperWDProperty(id=item.id, uri=item.get_uri(), aliases=item.aliases,
                                            label=item.label, description=item.description,
                                            parents=parents, parents_closure=parents_closure)

    @staticmethod
    def wd_value_deser(val: DataValue):
        if val.is_qnode():
            ent_id = val.as_qnode_id()
            if ent_id.startswith("Q"):
                uri = f"http://www.wikidata.org/entity/{ent_id}"
            elif ent_id.startswith("P"):
                uri = f"http://www.wikidata.org/prop/{ent_id}"
            else:
                uri = ent_id
            return Value(type=ValueType.URI, value=uri)
        if val.is_quantity():
            return Value(type=ValueType.Float, value=val.value['amount'])
        return Value(type=ValueType.String, value=val.to_string_repr())


def get_qnode_db(db_or_dbfile, read_only=False, proxy: bool=False):
    if isinstance(db_or_dbfile, (str, Path)):
        db = kg_db.get_qnode_db(db_or_dbfile, read_only=read_only, proxy=proxy)
    else:
        db = db_or_dbfile

    return StoreWrapper(db, GramsIntFn.key_deser, GramsIntFn.qnode_deser)


def get_ontclass_db(db_or_dbfile: str, read_only=False, proxy: bool=False):
    if isinstance(db_or_dbfile, (str, Path)):
        db = kg_db.get_wdclass_db(db_or_dbfile, read_only=read_only, proxy=proxy)
    else:
        db = db_or_dbfile

    return StoreWrapper(db, GramsIntFn.key_deser, GramsIntFn.ont_class_deser)


def get_ontprop_db(db_or_dbfile: str, read_only=False, proxy: bool=False):
    if isinstance(db_or_dbfile, (str, Path)):
        db = kg_db.get_wdprop_db(db_or_dbfile, read_only=read_only, proxy=proxy)
    else:
        db = db_or_dbfile

    return StoreWrapper(db, GramsIntFn.key_deser, GramsIntFn.ont_prop_deser)


def index_ontology(eshost: str):
    wdprops = WDProperty.from_file()
    docs = [
        {
            "id": wdprop.id,
            "uri": wdprop.get_uri(),
            "label": str(wdprop.label),
            "readable_label": f"{wdprop.label} ({wdprop.id})",
            "aliases": [str(x) for x in wdprop.aliases],
            "description": str(wdprop.description)
        }
        for wdprop in wdprops.values()
    ]
    docs.append({
        "id": str(RDFS.label),
        "uri": str(RDFS.label),
        "label": "rdfs:label",
        "readable_label": "rdfs:label",
        "aliases": ["label"],
        "description": ["label of a resource"]
    })
    OntologyPropertySearch(eshost).load2index(docs)

    wdclasses = WDClass.from_file()
    docs = [
        {
            "id": wdclass.id,
            "uri": wdclass.get_uri(),
            "label": str(wdclass.label),
            "readable_label": f"{wdclass.label} ({wdclass.id})",
            "aliases": [str(x) for x in wdclass.aliases],
            "description": str(wdclass.description)
        }
        for wdclass in wdclasses.values()
    ]
    docs.append({
        "id": "http://wikiba.se/ontology#Statement",
        "uri": "http://wikiba.se/ontology#Statement",
        "label": "wikibase:Statement",
        "readable_label": "wikibase:Statement",
        "aliases": ["statement"],
        "description": "Wikidata Statement"
    })
    OntologyClassSearch(eshost).load2index(docs)


if __name__ == '__main__':
    # db = get_qnode_db("/workspace/sm-dev/grams/data/qnodes.db", proxy=True)
    # url = 'http://www.w3.org/2000/01/rdf-schema#label'
    # print(url in db)
    # print(db.get(url, None))
    index_ontology("http://localhost:9200")


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sm_widgets',
 'sm_widgets.models',
 'sm_widgets.services',
 'sm_widgets.widgets',
 'sm_widgets.widgets.annotator',
 'sm_widgets_integration']

package_data = \
{'': ['*'], 'sm_widgets.widgets': ['jscodes/*']}

install_requires = \
['elasticsearch>=7.12.1,<8.0.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'ipycallback>=0.2.5,<0.3.0',
 'kgdata>=1.2.3,<2.0.0',
 'loguru>=0.5.3,<0.6.0',
 'orjson>=3.5.2,<4.0.0',
 'python-slugify>=5.0.2,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'sem-desc>=0.1.14,<0.2.0',
 'tqdm>=4.60.0,<5.0.0']

setup_kwargs = {
    'name': 'sm-widgets',
    'version': '0.1.13',
    'description': '',
    'long_description': None,
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List, TypedDict, Dict, Union, Tuple, Optional

import networkx as nx

from grams.algorithm.semantic_graph import SGStatementNode
from grams.algorithm.sm_wikidata import WDOnt
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import QNode, WDProperty
from sm_widgets.models import Link, Value, Table
from sm_widgets.models.entity import ValueType
from sm_widgets.widgets.annotator.assistant import AnnotatorAssistant, ColumnRelationshipResult, ColumnRelationship


def convert_table(table: LinkedTable):
    links = []
    for rlinks in table.links:
        links.append([
            [
                Link(start=link.start, end=link.end, url=link.url,
                     entity=f"http://www.wikidata.org/entity/{link.qnode_id}" if link.qnode_id is not None else None)
                for link in clinks
            ]
            for clinks in rlinks
        ])
    context_values = []
    if table.context.page_qnode is not None:
        uri = f"http://www.wikidata.org/entity/{table.context.page_qnode}"
        context_values.append(Value(ValueType.URI, value=uri))
    return Table(table=table.table, context_values=context_values, context_tree=[], links=links)


class GRAMSAnnotatorAssistant(AnnotatorAssistant):

    def __init__(self, inputs: List[TypedDict("input", table=LinkedTable, dg=nx.MultiDiGraph, sg=nx.MultiDiGraph)],
                 entities: Dict[str, QNode],
                 wdprops: Dict[str, WDProperty]):
        self.qnodes = entities
        self.wdprops = wdprops

        self.wdont = WDOnt(self.qnodes, {}, self.wdprops)

        self.table2g = {}
        for e in inputs:
            self.table2g[e['table'].id] = (e['dg'], e['sg'])

    def get_row_indices(self, table: LinkedTable, source_node: Union[int, str], target_node: Union[int, str],
                        links: Tuple[str, str]):
        assert isinstance(source_node, int) or isinstance(target_node,
                                                          int), "Can only get row index for at least one column"
        if isinstance(source_node, str):
            if table.context.page_qnode == source_node:
                # TODO: get context node id, is there any better way to do it?
                source_id = f"ent:{source_node}"
            else:
                source_id = f"ent:{source_node}"
        else:
            source_id = f"column-{source_node}"

        if isinstance(target_node, str):
            # TODO: fix me after we fix the data graph
            if table.context.page_qnode == target_node:
                # TODO: get context node id, is there any better way to do it?
                target_id = f"ent:{target_node}"
            else:
                target_id = f"ent:{target_node}"
        else:
            target_id = f"column-{target_node}"

        dg, sg = self.table2g[table.id]
        rows = set()
        for paths in nx.all_simple_edge_paths(sg, source_id, target_id, cutoff=2):
            assert len(paths) == 2
            uid, sid, eid = paths[0]
            _, vid, peid = paths[1]

            if not (eid == links[0] and peid == links[1]):
                continue

            stmt: SGStatementNode = sg.nodes[sid]['data']
            for (source_flow, target_flow), stmt2prov in stmt.flow.items():
                if target_flow.sg_target_id == vid and target_flow.edge_id == peid:
                    dv = dg.nodes[target_flow.dg_target_id]['data']
                    if dv.is_cell:
                        rows.add(dv.row)
                    else:
                        du = dg.nodes[source_flow.dg_source_id]['data']
                        assert du.is_cell
                        rows.add(du.row)
        return rows

    def get_column_relationships(self, table: LinkedTable, column_index: str) -> Optional[ColumnRelationshipResult]:
        dg, sg = self.table2g[table.id]
        node_id = f"column-{column_index}"

        incomings = []
        for sid, _, eid, edata in sg.in_edges(node_id, data=True, keys=True):
            # get statement
            s = sg.nodes[sid]['data']
            inedges = list(sg.in_edges(sid, data=True, keys=True))
            assert s.is_statement, f"Error in the semantic graph for column {column_index} (statement {sid})"
            assert len(inedges) == 1, f"Error in the semantic graph for column {column_index} (statement {sid})"
            uid, _, peid, pedata = inedges[0]

            freq = s.compute_freq(None, None, edata['data'], is_unique_freq=False)
            u = sg.nodes[uid]['data']
            if u.is_column:
                endpoint = u.column
            else:
                if u.qnode_id[0] == 'P':
                    # bug here
                    endpoint = {"uri": WDOnt.get_prop_uri(u.qnode_id), "label": self.wdont.get_pnode_label(u.qnode_id)}
                else:
                    endpoint = {"uri": WDOnt.get_qnode_uri(u.qnode_id), "label": self.wdont.get_qnode_label(u.qnode_id)}

            incomings.append(ColumnRelationship(
                endpoint=endpoint,
                predicates=(
                    {"uri": WDOnt.get_prop_uri(peid), "label": self.wdont.get_pnode_label(peid)},
                    {"uri": WDOnt.get_prop_uri(eid), "label": self.wdont.get_pnode_label(eid)},
                ),
                freq=freq
            ))

        outgoings = []
        for _, sid, peid, pedata in sg.out_edges(node_id, data=True, keys=True):
            # get statement
            s = sg.nodes[sid]['data']
            outedges = list(sg.out_edges(sid, data=True, keys=True))
            assert s.is_statement, f"Error in the semantic graph for column {column_index} (statement {sid})"

            for _, vid, eid, edata in outedges:
                freq = s.compute_freq(None, None, edata['data'], is_unique_freq=False)
                v = sg.nodes[vid]['data']
                if v.is_column:
                    endpoint = v.column
                else:
                    assert v.is_value
                    if v.is_entity_value:
                        if v.qnode_id[0] == 'P':
                            # bug here
                            endpoint = {"uri": WDOnt.get_prop_uri(v.qnode_id),
                                        "label": self.wdont.get_pnode_label(v.qnode_id)}
                        else:
                            endpoint = {"uri": WDOnt.get_qnode_uri(v.qnode_id),
                                        "label": self.wdont.get_qnode_label(v.qnode_id)}
                    else:
                        # not support yet
                        assert v.is_literal_value
                        continue

                outgoings.append(ColumnRelationship(
                    endpoint=endpoint,
                    predicates=(
                        {"uri": WDOnt.get_prop_uri(peid), "label": self.wdont.get_pnode_label(peid)},
                        {"uri": WDOnt.get_prop_uri(eid), "label": self.wdont.get_pnode_label(eid)},
                    ),
                    freq=freq
                ))

        return dict(incoming=incomings, outgoing=outgoings)


# if __name__ == '__main__':
#     from sm_unk.prelude import M
#     from sm_unk.dev.wikitable2wikidata.prelude import *
#     from sm_unk.config import *
#
#     dataset_dir = HOME_DIR / "wikitable2wikidata/500tables"
#
#     tables = [LinkedTable.from_dbpedia_table(Table.deser_str(r)) for r in
#               M.deserialize_lines(dataset_dir / "tables.jl.gz")]
#     tables = {x.table.metadata.table_id: x for x in tables}
#     tables = [tables[tbl_id] for tbl_id in M.deserialize_lines(dataset_dir / "predictions_order.txt", trim=True)]
#     tbl_context = M.deserialize_json(dataset_dir / "context.json.gz")
#
#     curated_model_index = [16]
#     annotate_data = []
#     assistant_data = []
#     for i in curated_model_index:
#         tbl = tables[i]
#         fname = tbl.get_friendly_fs_id()
#         context = tbl_context.get(tbl.id, [])
#
#         description = f"<b>table.index</b> = {i}. " + " > ".join(
#             [f"<b>[h{h['level']}]</b> {h['header'].strip()}" for h in context])
#         sg, dg = M.deserialize_pkl(dataset_dir / f"experiments/v_all/cache/init_sg/a{i:03d}_{fname}.pkl")[1:]
#         annotate_data.append((fname, description, tbl))
#         assistant_data.append({"table": tbl, "sg": sg, "dg": dg})
#
#     assistant = GRAMSAnnotatorAssistant(assistant_data, {}, {})
#     print(assistant.get_row_indices(tbl, 2, 'Q18239264', ('P39', 'P39')))



from typing import Optional

import orjson

import sm.outputs as O
from sm_widgets.models import Table
from sm_widgets.widgets.annotator.annotator import Annotator
from sm_widgets.widgets.annotator.session import Session


class SMVisualizer(Annotator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.app_js_render_fn = "Annotator.renderVizSemModelApp" if self.dev else "Annotator.renderDevVizSemModelApp"

    def visualize(self, id: str, note: str, sm1: O.SemanticModel, sm2: O.SemanticModel,
                  table: Optional[Table] = None, two_column_layout: bool = False):
        self.session = Session(id, True, note, table, [sm1, sm2])

        cmds = [
            {
                "type": "wait_for_client_ready",
            },
            {
                "type": "set_props",
                "props": {
                    "log": {
                        "note": note,
                    },
                    "graphs": [self.serialize_sm(graph) for graph in self.session.graphs],
                    "currentGraphIndex": 0,
                    "wdOntology": {},
                    "twoColumnLayout": two_column_layout,
                    "entities": {},
                    "assistant": {
                        "id": self.session.table.id
                    },
                }
            }
        ]

        if table is not None:
            cmds[1]['props']['table'] = self.serialize_table_schema()
            cmds.append({
                "type": "exec_func",
                "func": "app.tableFetchData",
                "args": []
            })
        else:
            self.table = None
        self.tunnel.send_msg(orjson.dumps(cmds).decode())



from collections import Counter
from typing import List

from sm_widgets.models import Table
from sm.outputs import SemanticModel

class Session:

    def __init__(self, id: str, is_curated: bool, note: str, table: Table, graphs: List[SemanticModel]):
        self.id = id
        self.is_curated = is_curated
        self.note = note
        self.table = table
        self.graphs = graphs

        # use it for querying records
        self.table_records_index = set(range(table.size()))
        self.column2name = {}
        cname2freq = Counter(c.name for c in table.table.columns)
        for c in table.table.columns:
            if cname2freq[c.name] > 1:
                self.column2name[c.index] = f"{c.name} ({c.index})"
            else:
                self.column2name[c.index] = c.name

    def to_json(self):
        return {
            "version": 2,
            "table_id": self.table.id,
            "semantic_models": [sm.to_dict() for sm in self.graphs],
            "is_curated": self.is_curated,
            "note": self.note,
        }